// EffectFactory.cpp: implementation of the CEffectFactory class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "EffectFactory.h"
#include "PanFactory.h"
#include <wtypes.h>
#include "StringProp.h"
#include ".\\resource.h"
#include "UnregisterServer.h"
#include "PanUtils.h"
#include <vector>
using namespace std;

typedef enum
{
  ByProgID = 0,
  ByCLSID  = 1
} ComponentName;

//
// A word or two about how we persist Pan effects:
//
// OLE's structured storage (and streams) are used to hold the
// settings. The storage uses the following streams:
//
//        PanImage   - holds the UI control values
//        UISettings - holds the viewer persisted settings.
//
// PanImage has got the following structure:
//
//       kind     -- tag which specifies how the component is named:
//                   e.g., by CLSID or ProgID.
//       component name / CLSID.
//       value+   -- one or more values persisting the UI settings.
//
// UISettings has got the following structure:
//
//       string+  -- one or more strings (all of which are length prefixed, btw)
//
//  The strings are used to hold what images were imported at the time of the
//  save & for text input controls, the text typed in. This can (and will)
//  be extended to also hold font info + other viewer settings such as zoom & pan.
//


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CEffectFactory::CEffectFactory(int *pDisplayPasses, double *pMaxJitter, CComBSTR docURL) 
  : m_hFont(NULL), m_hWndMain(NULL), m_pDisplayPasses(pDisplayPasses),
    m_pMaxJitter(pMaxJitter), m_docURL(docURL)
{

}

CEffectFactory::~CEffectFactory()
{
    DeleteObject(m_hFont);
}

//////////////////////////////////////////////////////////////////////
// Local helper functions
//////////////////////////////////////////////////////////////////////



//
// FUNCTION: HRESULT LoadFromProgID(IStream*, REFIID, void**)
//
// PURPOSE:  Behaves like OleLoadFromStream(), but instead of prefixing
//           the stream with the CLSID of a component, it's ProgID is.
//           The ProgID string is length prefixed.
//
// COMMENTS: Naming a component by ProgID rather than CLSID, allows us
//           to refer to the component by name, i.e., its particular 
//           implementation (=> CLSID) might change without affecting
//           us.
static
HRESULT
LoadFromProgID(IStream*                pStream,
               LPWSTR*                 pProgID,
               REFIID                  iidInterface,
               CComQIPtr<IPanImageUI>& pPanUI
               )
{
    USES_CONVERSION;
    ULONG           ulLength  = 0;
    ULONG           cbRead    = 0;
    LPWSTR          progIDStr = NULL;

    // Fetch out the ProgID from the stream.
    pStream->Read(&ulLength, sizeof(ULONG), &cbRead);
    
    if (ulLength > 0) {
        progIDStr = (LPWSTR)malloc((ulLength+1) * sizeof(WCHAR));

        if (progIDStr) {
            pStream->Read(progIDStr,sizeof(WCHAR)*ulLength,&cbRead);
            progIDStr[ulLength] = L'\0';
            *pProgID = progIDStr;
        }
    }

    if (progIDStr) {
        // Instantiate the component given the ProgID.
        BIND_OPTS bOpts;
        ZeroMemory(&bOpts,sizeof(BIND_OPTS));
        bOpts.cbStruct = sizeof(BIND_OPTS);
        bOpts.grfFlags = BIND_MAYBOTHERUSER;
        bOpts.grfMode  = STGM_READWRITE;
        bOpts.dwTickCountDeadline = 0;

        IPersistStream* pPStream;
        HRESULT hr;
        CLSID  clsid;

        hr = CLSIDFromProgID(progIDStr, &clsid);

        if (SUCCEEDED(hr)) {
            hr = pPanUI.CoCreateInstance(clsid);
        } else {
            // something's missing - either the ProgID entry in the Reg. or the component.
            // Notify the user that something unexpected occurred.
            //
            // The message string is assumed to contain an insert for the ProgID string.
            // 
            TCHAR  msgString[1024];
            TCHAR* msgBuf;
            LPVOID lppArgs[1];

            // drop the "Pan." prefix
            lppArgs[0] = (void*)wcschr(progIDStr,(int)L'.');
            if (lppArgs[0] != NULL) {
                lppArgs[0] = (void*) ((WCHAR*)lppArgs[0] + 1);
            }
            ::LoadString(NULL, IDS_NO_COMPONENT, msgString, 1024);
            // Initial attempt used FORMAT_MESSAGE_FROM_HMODULE, but it 
            // requires the storage of strings in MESSAGETABLEs which can only
            // be produced, I believe, by a message language & compiler. Thanks,
            // but no thanks.
            FormatMessage( FORMAT_MESSAGE_FROM_STRING | 
                           FORMAT_MESSAGE_ALLOCATE_BUFFER |
                           FORMAT_MESSAGE_ARGUMENT_ARRAY,
                           msgString,
                           0,
                           0,
                           (LPTSTR)&msgBuf,
                           1024, // what's the purpose of this arg in this mode?
                           (char**)lppArgs);
            ::MessageBox(NULL, msgBuf, TEXT("Pan image viewer"), MB_OK | MB_ICONERROR);
            LocalFree((HLOCAL)msgBuf);
            return hr;
        }

        if (SUCCEEDED(hr)) {
            // Call IPersistStream->Load() on the instantiated
            // component.
            hr = pPanUI->QueryInterface(&pPStream);

            if (SUCCEEDED(hr)) {
                hr = pPStream->Load(pStream);
            }
        } else {
            // failed to instantiate the component - inform the user.
            // the ProgID either didn't exist, the component didn't (anymore?)
            // Notify the user that something unexpected occurred.
            TCHAR msgString[1024];
            TCHAR* msgBuf;
            LPVOID lppArgs[1];

            // drop the Pan. prefix
            lppArgs[0] = (void*)wcschr(progIDStr,(int)L'.');
            if (lppArgs[0] != NULL) {
                lppArgs[0] = (void*) ((WCHAR*)lppArgs[0] + 1);
            }
            ::LoadString(NULL, IDS_NOT_CREATED, msgString, 1024);
            FormatMessage( FORMAT_MESSAGE_FROM_STRING | 
                           FORMAT_MESSAGE_ALLOCATE_BUFFER |
                           FORMAT_MESSAGE_ARGUMENT_ARRAY,
                           msgString,
                           0,
                           0,
                           (LPTSTR)&msgBuf,
                           1024, // what's the purpose of this arg in this mode?
                           (char**)lppArgs);
            ::MessageBox(NULL, msgBuf, TEXT("Pan image viewer"), MB_OK | MB_ICONERROR);
            
        }
        return hr;

    } else {
        return E_FAIL; // how informative
    }

}

//
// FUNCTION: GetEffectName(REFCLSID rclsid, LPCTSTR&)
//
// PURPOSE:  Given the CLSID of a Pan component, look up it's
//           user-friendly name.
//
static
HRESULT
GetEffectName(REFCLSID rclsid, LPCTSTR& effectName)
{
    USES_CONVERSION;
    HKEY        hKey;
    TCHAR       subKeyStr[1024];
    LONG        lRes;
    OLECHAR     clsidString[41];
    HRESULT     hr;

    hr = StringFromGUID2(rclsid,clsidString,40);

    if (SUCCEEDED(hr)) {

        _stprintf(subKeyStr,TEXT("CLSID\\%s"), OLE2T(clsidString));

        lRes = RegOpenKeyEx(HKEY_CLASSES_ROOT,
                            subKeyStr,
                            0,
                            KEY_READ,
                            &hKey);
        if ( lRes == ERROR_SUCCESS ) {
            DWORD dwType;
            DWORD dwLen = 0;

            lRes = RegQueryValueEx(hKey, NULL, NULL, &dwType, NULL, &dwLen);
            if ( lRes == ERROR_SUCCESS && dwType == REG_SZ ) {
                // Got the length, now go and grab the string.
                effectName = new TCHAR [dwLen + 1];
                lRes = RegQueryValueEx(hKey, NULL, NULL, &dwType, (LPBYTE)effectName, &dwLen);
            }
            RegCloseKey(hKey);
            return (HRESULT_FROM_WIN32(lRes));
        } else {
            return HRESULT_FROM_WIN32(lRes);
        }
    } else {
        return hr;
    }

}

//////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////

//
// FUNCTION: CEffect* OpenEffect(HWND, REFCLSID)
//
// PURPOSE:  Given the CLSID of a component, instantiate the Pan
//           component and its UI. Used by File->New
//
CEffect*
CEffectFactory::OpenEffect(HWND hWndParent, REFCLSID clsid, PTSTR progID, CPerfData* pData)
{
    CComQIPtr<IPanImage>   img;
    CComQIPtr<IPanImageUI> pUI;
    CWindowVirt*           winUI;
    CPanFactory*           pFact;
    CViewWindow*           pView;
    HRESULT                hr;

    vector<CStringProp*>* propVec = new vector<CStringProp*>();
    vector<CFontProp*>* fontVec   = new vector<CFontProp*>();

    hr = pUI.CoCreateInstance(clsid);

    if (SUCCEEDED(hr)) {
        pUI->QueryInterface(&img);
        
        pFact = new CPanFactory(img,m_pDisplayPasses,m_pMaxJitter, m_docURL);

        winUI = pFact->Create(hWndParent,propVec,fontVec,pData,&pView);

        // Use the same font throughout the UI.
        HDC hDCUI = GetDC(NULL);

        if (!m_hFont) {
           m_hFont =
            CreateFont(-MulDiv(8, GetDeviceCaps(hDCUI, LOGPIXELSY), 72),
                       0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, VARIABLE_PITCH | FF_SWISS, TEXT(""));
        }
        CPanUtils::BroadcastFont(*winUI, m_hFont);
        ReleaseDC(NULL,hDCUI);

        CEffect* ef = new CEffect(winUI, img, propVec, fontVec, progID, pView, m_hWndMain);

        delete pFact;

        return ef;
    } else {
        // Failed to instantiate the component, offer the user the option of removing
        // the component from the list of registered effects.
        TCHAR msgBuf[1024];
        ::LoadString(NULL, IDS_NO_EFFECT, msgBuf, 1024);
        if (::MessageBox(NULL, msgBuf, TEXT("Pan image viewer"), MB_YESNO | MB_ICONQUESTION) == IDYES) {
            UnregisterServer(clsid);
        }
        return NULL;
    }
}

//
// FUNCTION:    CEffect* OpenEffect(HWND, LPCTSTR)
//
// PURPOSE:     Given the name of a Pan settings file, attempt
//              to instantiate the Pan component it refers to +
//              configure the effect (and its UI) to hold the values
//              contained in the settings file. Used by File->Open.
//
// COMMENTS:    See comments at the top of this file regarding details
//              of how we persist Pan components.
//
CEffect*
CEffectFactory::OpenEffect(HWND       hWndParent,
                           LPCTSTR    fName,
                           CPerfData* pData,
                           LPTSTR&    pEffectName
                          )
{
    USES_CONVERSION;
    CPanFactory*            pFact;
    CWindowVirt*            winUI;
    IStorage*               pStg;
    IPersist*               pPersist;
    IStream*                pStream;
    CComQIPtr<IPanImageUI>  pUI;
    CComQIPtr<IPanImage>    img;

    HRESULT hr;
    vector<CStringProp*>* propVec = new vector<CStringProp*>();
    vector<CFontProp*>* fontVec   = new vector<CFontProp*>();
    CViewWindow* pView;

    LPCWSTR fNameW = T2CW(fName);
    unsigned char   cKind;
    ULONG           ulRead;
    LPWSTR          tProgID = NULL;

    hr = 
    StgOpenStorage(fNameW,
                   NULL,
                   STGM_READ | STGM_SHARE_EXCLUSIVE,
                   NULL,0, &pStg);
#if 0
        // StgOpenStorageEx is only supported on Win2k and later
      StgOpenStorageEx(fNameW,
                       STGM_READ | STGM_SHARE_EXCLUSIVE,
                       STGFMT_DOCFILE,
                       0,
                       NULL,
                       0,
                       IID_IStorage,
                       reinterpret_cast<void**>(&pStm));
#endif

    if (FAILED(hr)) {
        return NULL;
    }

    // Instantiate the component and set the values associated with
    // its UI controls.
    if ((hr = pStg->OpenStream(L"PanImage", NULL, 
                           STGM_READ | STGM_SHARE_EXCLUSIVE,
                           0,&pStream)) == S_OK) {
        pStream->Read(&cKind,1,&ulRead);
        if (ulRead != 1) {
            pStream->Release();
            return NULL;
        }
        switch ((ComponentName)cKind) {
        case ByProgID:
            hr = LoadFromProgID(pStream, &tProgID, __uuidof(pUI), pUI);
            break;
        case ByCLSID:
            hr = OleLoadFromStream(pStream, __uuidof(pUI), reinterpret_cast<void**>(&pUI));
        }
        pStream->Release();
    }
 
    if (SUCCEEDED(hr)) {
        pUI->QueryInterface(&img);
        
        pFact = new CPanFactory(img,m_pDisplayPasses,m_pMaxJitter, m_docURL);
        winUI = pFact->Create(hWndParent,propVec,fontVec,pData,&pView);

        int          strLen;
        DWORD        cbRead;
        WCHAR        tempStr[1024];
        int          vecSize = propVec->size();
        CStringProp* prop;

        if ((hr = pStg->OpenStream(L"UISettings", NULL, 
                                   STGM_READ | STGM_SHARE_EXCLUSIVE,
                                   0,&pStream)) == S_OK) {

            for (int i=0; i < vecSize; i++) {
                pStream->Read(&strLen,4,&cbRead);
                pStream->Read(tempStr,sizeof(WCHAR)*(strLen+1),&cbRead);
                prop = (*propVec)[i];
                prop->SetString(W2T(tempStr));
            }
            pStream->Release();
        }

        if ((hr = pStg->OpenStream(L"FontSettings", NULL,
                                   STGM_READ | STGM_SHARE_EXCLUSIVE,
                                   0,&pStream)) == S_OK) {
            FontInfo fi;
            vecSize = fontVec->size();

            for (int i=0;i < vecSize; i++) {
                pStream->Read(&fi,sizeof(FontInfo),&cbRead);
                (*fontVec)[i]->SetFontInfo(fi);
            }
            pStream->Release();
        }

        if ((hr = pStg->OpenStream(L"MouseInput", NULL,
                                   STGM_READ | STGM_SHARE_EXCLUSIVE,
                                   0,&pStream)) == S_OK) {
            struct {float x,y;} pos;
            if ( SUCCEEDED(pStream->Read(&pos,sizeof(pos),&cbRead)) ) {
                pFact->UpdateMousePos(pos.x,pos.y);
            }
            pStream->Release();
        }

        // A little bit controversial to resize the entire UI upon
        // loading a settings file, but restoring view size has its
        // uses.
        if ((hr = pStg->OpenStream(L"ViewSize", NULL,
                                   STGM_READ | STGM_SHARE_EXCLUSIVE,
                                   0,&pStream)) == S_OK) {
            SIZE sz;
            if ( SUCCEEDED(pStream->Read(&sz,sizeof(SIZE),&cbRead)) ) {
                SetWindowPos(m_hWndMain,NULL,0,0,sz.cx,sz.cy, SWP_NOMOVE | SWP_NOZORDER | SWP_NOCOPYBITS);
            }
            pStream->Release();
        }


        pStg->Release();

        // Use the same font throughout the UI.
        if (!m_hFont) {
            m_hFont = 
              CreateFont(-MulDiv(8, GetDeviceCaps(GetDC(NULL), LOGPIXELSY), 72),
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, VARIABLE_PITCH | FF_SWISS, TEXT(""));
        }
        CPanUtils::BroadcastFont(*winUI, m_hFont);

        hr = pUI->QueryInterface(&pPersist);
        if (SUCCEEDED(hr)) {
            CLSID clsid;
            pPersist->GetClassID(&clsid);
            GetEffectName(clsid,pEffectName);
        }

        return (new CEffect(winUI, img, propVec, fontVec, W2T(tProgID), pView, m_hWndMain));
    } else {
        delete propVec;
        delete fontVec;
        return NULL;
    }
}


#if 0 // not needed here, but useful.

//
// FUNCTION: RegisterCategory
//
// PURPOSE: Register a new component category on the local machine.
//          Abstracts away from the details of having to deal with
//          Component Categories Manager.
//
static
HRESULT
RegisterCategory(REFIID catid, LPCOLESTR szDesc)
{
    CATEGORYINFO cI;
    ICatRegister *pci = 0;

    cI.catid = catid;
    // Note: the locale for the descriptive string of the category
    // is fixed to US-English.
    cI.lcid  = MAKELCID(MAKELANGID(LANG_ENGLISH,SUBLANG_ENGLISH_US), SORT_DEFAULT);
    wcsncpy(cI.szDescription,szDesc, 127);

    HRESULT hr = 
        CoCreateInstance
            ( CLSID_StdComponentCategoriesMgr, 0, 
              CLSCTX_ALL, IID_ICatRegister,
              reinterpret_cast<void**>(&pci));

    if (SUCCEEDED(hr)) {
        hr = pci->RegisterCategories(1,&cI);
        pci->Release();
    }

    return hr;
}

#endif
