/*
 * To specify the UI that you want the GUI library to instantiate,
 * you describe the hierarchy via a UI struct value -- see uiSpec.cpp
 * for more comments as to what (and why) these functions do.
 *
 */
#ifndef __UISPEC_H__
#define __UISPEC_H__

#include "PanDisp.h"

//
// Kinds of UI elements supported:
//

typedef enum 
	{ Slider
	, ISlider
	, Radio
	, Tick
	, MouseInput
	, Button
	, Label
	, ImportImage
	, VBox
	, HBox
	, Display
	, PhSDisplay
	, Resizeable
	, TextBox
	} UIKind;

typedef struct VBoxInfo {
  char*      label;
  int        noChildren;
  struct UI* children[1];
} VBoxInfo;

typedef struct HBoxInfo {
  char*         label;
  int           noChildren;
  struct UI*    children[1];
} HBoxInfo;

typedef struct SliderInfo {
  char*           label;
  float           min;
  float           max;
  float           initVal;
  FloatUpdater*   pUpdateValue;
  void*           pParam;
} SliderInfo;

typedef struct ISliderInfo {
  char*           label;
  int             min;
  int             max;
  int             initVal;
  IntUpdater*     pUpdateValue;
  void*           pParam;
} ISliderInfo;

typedef struct TickInfo {
  char*           label;
  int             defValue;
  IntUpdater*     pUpdateValue;
  void*           pParam;
} TickInfo;

typedef struct ButtonInfo {
  char*  label;
  int    value;
} ButtonInfo;

typedef struct LabelInfo {
  char*  label;
} LabelInfo;

typedef struct MouseInputInfo {
  FloatUpdater*   updMouse;
  void*           pParamX;
  void*           pParamY;
} MouseInputInfo;

typedef struct RadioInfo {
  char*           label;
  int             defValue;
  IntUpdater*     pUpdateValue;
  void*           pParam;
  int             noItems;
  char*           labels[1];
} RadioInfo;

typedef struct ImageInfo {
  char*           label;
  char*           defImg;
  IntUpdater*     pUpdateValue;
  void*           pParamBits;
  void*           pParamWidth;
  void*           pParamHeight;
} ImageInfo;

typedef struct TextInfo {
  char*           label;
  char*           defVal;
  IntUpdater*     pUpdateValue;
  void*           pParamBits;
  void*           pParamWidth;
  void*           pParamHeight;
} TextInfo;

typedef struct ResizeInfo {
  BOOL hResize;
  BOOL vResize;
  struct UI* uiElem;
} ResizeInfo;

typedef struct DisplayInfo {
  PanImageDisplay* displayer;
  BOOL	     timeVarying;
  int	     nWidth;
  int        nHeight;
} DisplayInfo;

typedef struct PhSDisplayInfo {
  void*      pImg;
  int	     nWidth;
  int        nHeight;
} PhSDisplayInfo;

// 
// The 'sum' of different constructors/elements is repr. by
// the UI struct, which stores a tag along with the struct
// describing that element.
// 
typedef union {
  SliderInfo*     slider;
  ISliderInfo*    islider;
  RadioInfo*      radio;
  ImageInfo*      image;
  HBoxInfo*       hbox;
  VBoxInfo*       vbox;
  TickInfo*       tick;
  MouseInputInfo* mouseInput;
  ButtonInfo*     button;
  LabelInfo*      label;
  ResizeInfo*     resize;
  DisplayInfo*    display;
  PhSDisplayInfo* phsdisplay;
  TextInfo*       textBoxInfo;
} UIInfo;

typedef struct UI {
  UIKind kind;
  UIInfo info;
} UI;

//
// UI constructors
// 
extern HWND createUI(HWND hWndParent, UI* pSpec);
extern void freeUI(UI* ui);
extern "C" int LoadImportedImg(char* fName, int* w, int* h);

extern UI* mkVBox(char* lab, int no, ...);
extern UI* mkHBox(char* lab, int no, ...);
extern UI* mkSlider(char* lab,
		    float min,
		    float max,
		    float initVal,
		    FloatUpdater* pUpd,
		    void* pParam);

extern UI* mkISlider(char* lab,
		     int min,
		     int max,
		     int initVal,
		     IntUpdater* pUpd,
		     void* pParam);
extern UI* mkTick(char* lab,
		  int defVal,
		  IntUpdater* pUpd,
		  void* pParam);
extern UI* mkMouseInput(FloatUpdater* pUpd, void* pParamX, void* pParamY);
extern UI* mkButton(char* lab, int val);
extern UI* mkLabel(char* lab);
extern UI* mkImportImage(char* lab,
			 char* defImg,
			 IntUpdater* pUpd,
			 void* pParamBits,
			 void* pParamWidth,
			 void* pParamHeight);
extern UI* mkRadio(char* lab,
		   int defVal,
		   IntUpdater* pUpd,
		   void* pParam,
		   int no,
		   ...);
extern UI* mkResizeable(BOOL hResiz, BOOL vResiz, UI* elem);
extern UI* mkViewWindow(PanImageDisplay* pImg, BOOL tVarying, int nWidth, int nHeight);
extern UI* mkPhSViewWindow(void* pImg, int nWidth, int nHeight);
extern UI* mkTextBox(char* lab,
		     char* defVal,
		     IntUpdater* pUpd,
		     void* pParamBits,
		     void* pParamWidth,
		     void* pParamHeight);

#endif
