-- Expression type with normalization during construction.
--
-- Dynamically typed version.  See also Exp.hs, which provides statically
-- typed wrappers.
-- 
-- As in Fran, we overload where possible.  When the types don't allow it
-- (e.g., "Bool" is hard-wired into a method signature), we append "E" to
-- a name if alphanumeric, or "*" if not.  Examples: "floorE" and "==*".
-- If we were to suppress some of the Prelude classes, we might be able to
-- replace them with more general versions using multi-parameter type
-- classes.
--
-- To do:
-- + Split this module into (a) basic DExp stuff (incl hnf, etc) and (b)
--   "smart constructors".  What's now in DExpUtils would go in (a).
-- + Replace Deref8, Deref32 with ReadArr
-- + Get the fromBF/toBF simplifications working
-- + Front-end compiler, that turns nice rule specification into the form
--   used here.  May be a low priority.


module DExp where

import Maybe ( fromMaybe )
import List (nub)
import IOExts
import Memo

import BasicTypes
import Types

-- Expression type.  Some of these constructors do not show up in the normal
-- form (NF), as marked.  They're here only for convenience of
-- pretty-printing or code generation.

-- "Let" is there just for a final CSE phase.  We're using the source
-- language's "let" directly.  The main reason is for implementation
-- simplicity, since as far as I can tell, we want very aggressive
-- inlining.  Since normalization is done bottom-up, during construction,
-- inlining does not lead to redundant simplification.  Another benefit is
-- pretty Haskell definitions for the image algebra and sample
-- applications.  If we end up layering another language on top even for
-- these definitions, then this prettiness won't matter much.

data Exp =
    -- literals
    LitInt Int | LitDouble Double | LitBool Bool -- | LitF32vec4 V4.F32vec4
  | ImportData String           -- see note below

  | Var Id Type                 -- variables
  | Lambda Id Type Exp          -- Function

  -- Binding constructs.  ToDo: redefine all in terms of Lambda
  | Let Id Exp Exp              -- binding
  | If Exp Exp Exp              -- conditional
  | Loop Id Exp Exp             -- iteration var, limit, body
  | NewArr Exp Type
  | MkArr Exp Id Exp            -- size, index name, element exp
  -- These ones do use Lambda
  | NewRef Exp                  -- initial value
  | ReadRef Exp                 -- ref
  | WriteRef Exp Exp            -- ref, new val exp.

  -- Non-binding operators
  | Add Exp Exp | Negate Exp | Sub Exp Exp  -- Sub is not in NF
  | Mul Exp Exp | Recip Exp | Divide Exp Exp  -- Divide is not in NF
  | Sqrt Exp  -- not in NF
  | Sin Exp | Cos Exp | Atan2 Exp Exp
  | LogBase Exp Exp | Pow Exp Exp   -- logBase and (**) (note arg order)
  | I2f Exp | F2i Exp | B2i Exp | Abs Exp
  | Mod Exp Exp | Div Exp Exp
  | Equal Exp Exp | LessThan Exp Exp | Min Exp Exp | Max Exp Exp
  | And Exp Exp | Or Exp Exp | Not Exp
  | LShift Exp Exp | BitAnd Exp Exp | BitOr Exp Exp | BitNot Exp
  | RShift Exp Exp  -- Not in normal form, but used in compilation
  | PackF32 Exp Exp Exp Exp | NthF32 Exp Exp | SumF32 Exp
  | MkTuple [Exp] | TupleElt Int Exp -- index (>= 0) then tuple
  | Deref8 Exp
  | Deref32 Exp
  | Noop
  | SetInt Exp Exp        -- addr, value
     -- array operations:
  | ReadArr Exp Exp       -- arr, index
  | ReadMutArr Exp Exp    -- arr, index
  | WriteArr Exp Exp Exp  -- arr, index, value
  
  | Rand

  | Bind Exp Exp   -- monadic bind, i.e., the second arg is assumed to 
  		   -- be of the form (Lambda v t e)
  | Return Exp     -- monadic unit op.
  | Perform Exp
  
    -- backend specific primitives
  | DoProgress  Exp{-current value-} Exp{-max value-}
  | Copy1DArray Exp{-base pointer-}  Exp{-length-}
 deriving (Eq, Show)


-- Fills in for IO ().  No values are created, however.
data Action a = MkAction deriving (Show, Eq)

-- Note: if you extend the Exp type, be sure also to extend the lists in
-- CCompile.hs and DExpUtils.hs

{-
 Re: ImportData constructor : (ImportData String)

 The ImportData constructor is used to represent an int value
 which becomes bound/is imported at run-time. The String value 
 specifies the location where this run-time bound value can be
 found.
 
 One (and, currently, only) use of the ImportData constructor is to
 do bitmap importation in stand-alone executables/plugins, with the
 String specifying the path to the BMP image to import.
 
 The transformational stages of the compiler will hoist occurrences of
 ImportData to the very top (we hope). See CCompile for an example of
 how ImportData is converted into running code.
-}

typeOf :: Exp -> Type

typeOf LitInt{}         = Int
typeOf LitDouble{}      = Double
typeOf LitBool{}        = Bool
typeOf ImportData{}     = Int
typeOf (Var _ ty)       = ty
typeOf (Lambda x ty e)  = Fun ty (typeOf e)
typeOf (Let _ _  e)     = typeOf e
typeOf (If c a b)       = typeOf a
typeOf (Add e e')       = typeOf e
typeOf (Negate e)       = typeOf e
typeOf (Mul e e')       = typeOf e
typeOf (Recip e)        = typeOf e
typeOf (Pow e e')       = Double
typeOf (LogBase e e')   = Double
typeOf Sin{}            = Double
typeOf Cos{}            = Double
typeOf Atan2{}          = Double
typeOf I2f{}            = Double
typeOf F2i{}            = Int
typeOf B2i{}            = Int
typeOf (Abs e)          = typeOf e
typeOf Mod{}            = Int
typeOf Div{}            = Int
typeOf Equal{}          = Bool
typeOf LessThan{}       = Bool
typeOf (Min e _)        = typeOf e
typeOf (Max e _)        = typeOf e
typeOf And{}            = Bool
typeOf Or{}             = Bool
typeOf Not{}            = Bool
typeOf LShift{}         = Int
typeOf RShift{}         = Int
typeOf BitAnd{}         = Int
typeOf BitOr{}          = Int
typeOf BitNot{}         = Int
typeOf PackF32{}        = F32vec4
typeOf NthF32{}         = Double
typeOf SumF32{}         = Double
typeOf (MkTuple elts)   = Tuple (map typeOf elts)
typeOf (TupleElt n tup) = tys!!n where Tuple tys = typeOf tup
typeOf Deref8{}         = Int
typeOf Deref32{}        = Int
typeOf SetInt{}         = Action Int
typeOf Noop             = Action Void
typeOf Loop{}           = Action Void
typeOf (Bind a (Lambda _ _ y)) = typeOf y
typeOf (Bind a (Let _ _ e)) = typeOf (Bind a e)
typeOf e@Bind{}           = error ("typeOf.Bind: Expected a Lambda on the right: " ++ show e)
typeOf (Return e)       = Action (typeOf e)
typeOf (Perform e)      = getActionType (typeOf e)

typeOf e@(ReadMutArr arr _)= Action ty where Arr ty = typeOf arr
typeOf e@(ReadArr arr _) = ty where Arr ty = typeOf arr
typeOf WriteArr{}       = Action Void
typeOf (NewArr _ ty)    = Action (Arr ty)
typeOf (MkArr _ _ elem) = Arr (typeOf elem)
typeOf Rand             = Action Int
typeOf (NewRef init)    = Action (Ref (typeOf init))
typeOf (ReadRef r)      = Action (getRefType (typeOf r))
typeOf (WriteRef r v)   = Action Void
typeOf DoProgress{}     = Action Void
typeOf Copy1DArray{}    = Action Void

typeOf e = error ("DExp.typeOf: no case for " ++ show e)


-- Is a literal.  Sometimes simplifies rewrite rules.
isLit :: Exp -> Bool
isLit LitInt{}    = True
isLit LitDouble{} = True
isLit (PackF32 LitDouble{} LitDouble{} LitDouble{} LitDouble{}) = True
isLit _ = False


-- So simple that there's no need to avoid recomputing
isSimple               :: Exp -> Bool
isSimple LitInt{}      = True
isSimple LitDouble{}   = True
isSimple LitBool{}     = True
isSimple ImportData{}  = True  -- specially handled in back-end (for now)
isSimple Var{}         = True
isSimple DoProgress{}  = True
isSimple Copy1DArray{} = True
isSimple _             = False


-- Has a free occurrence somewhere in expression.
occFreeIn :: Id -> Exp -> Bool
occFreeIn name = occ
 where
   occ LitInt{}          = False
   occ LitDouble{}       = False
   occ LitBool{}         = False
   occ ImportData{}      = False
   occ (Var n _)           = name == n
   occ (Lambda x ty e)     = name /= x && occ e
   -- Hm.. "If" should probably have a hnf also.  It can still
   -- be treated specially when necessary.
   occ (If e1 e2 e3)       = occ e1 || occ e2 || occ e3
   occ (Let x e1 e2)       = occ e1 || (name /= x && occ e2)
   occ (Loop x e1 e2)      = occ e1 || (name /= x && occ e2)
   occ (NewArr e _)        = occ e
   occ (MkArr e1 x e2)     = occ e1 || (name /= x && occ e2)
   occ appl                = or (map occ args)
                              where (head,args) = toHnf appl



----- Identifier generation 

-- Sometimes it's necessary to create a new identifier, e.g., during CSE
-- and for index variables.

newId' = \ prefix -> do
  curr <- readIORef idRef
  writeIORef idRef (curr + 1)
  let name = prefix ++ show curr
  --putStrLn $ "newId': Generated " ++ name
  return name
 where
   idRef :: IORef Int
   idRef = unsafePerformIO (newIORef 0)


-- Head normal form support.  For generic operations, like code generation
-- and CSE.  Clients should handle all non-applications before using
-- toHnf.

data Op =
    AddOp | NegateOp | SubOp
  | MulOp | RecipOp | DivideOp | SqrtOp
  | SinOp | CosOp | Atan2Op | LogBaseOp | PowOp
  | I2fOp | F2iOp | B2iOp | AbsOp
  | ModOp | DivOp
  | EqualOp | LessThanOp | MinOp | MaxOp
  | AndOp | OrOp | NotOp
  | LShiftOp | BitAndOp | BitOrOp | BitNotOp
  | RShiftOp
  | PackF32Op | NthF32Op | SumF32Op
  | MkTupleOp | TupleEltOp Int
  | SetIntOp | Deref8Op | Deref32Op | NoopOp
  | ReadArrOp Type | ReadMutArrOp Type | WriteArrOp Type | NewArrOp Type
  | RandOp
  | BindOp | ReturnOp | PerformOp
  | NewRefOp | ReadRefOp | WriteRefOp
  | DoProgressOp | Copy1DArrayOp
  | IfOp   -- Phasing in
 deriving (Eq, Ord, Show)


toHnf :: Exp -> (Op, [Exp])

toHnf (Add e e')            = (AddOp, [e,e'])
toHnf (Negate e)            = (NegateOp, [e])
toHnf (Sub e e')            = (SubOp, [e,e'])
toHnf (Mul e e')            = (MulOp, [e,e'])
toHnf (Recip e)             = (RecipOp, [e])
toHnf (Divide e e')         = (DivideOp, [e,e'])
toHnf (Sin e)               = (SinOp, [e])
toHnf (Cos e)               = (CosOp, [e])
toHnf (Atan2 e e')          = (Atan2Op, [e,e'])
toHnf (Pow e e')            = (PowOp, [e,e'])
toHnf (LogBase e e')        = (LogBaseOp, [e,e'])
toHnf (Sqrt e)              = (SqrtOp, [e])
toHnf (I2f e)               = (I2fOp, [e])
toHnf (F2i e)               = (F2iOp, [e])
toHnf (B2i e)               = (B2iOp, [e])
toHnf (Abs e)               = (AbsOp, [e])
toHnf (Mod e e')            = (ModOp, [e,e'])
toHnf (Div e e')            = (DivOp, [e,e'])
toHnf (Equal e e')          = (EqualOp, [e,e'])
toHnf (LessThan e e')       = (LessThanOp, [e,e'])
toHnf (Min e e')            = (MinOp, [e,e'])
toHnf (Max e e')            = (MaxOp, [e,e'])
toHnf (And e e')            = (AndOp, [e,e'])
toHnf (Or e e')             = (OrOp, [e,e'])
toHnf (Not e)               = (NotOp, [e])
toHnf (LShift e e')         = (LShiftOp, [e,e'])
toHnf (RShift e e')         = (RShiftOp, [e,e'])
toHnf (BitAnd e e')         = (BitAndOp, [e,e'])
toHnf (BitOr e e')          = (BitOrOp, [e,e'])
toHnf (BitNot e)            = (BitNotOp, [e])
toHnf (PackF32 e1 e2 e3 e4) = (PackF32Op, [e1, e2, e3, e4])
toHnf (NthF32 e e')         = (NthF32Op, [e,e'])
toHnf (SumF32 e)            = (SumF32Op, [e])
toHnf (MkTuple es)          = (MkTupleOp, es)
toHnf (TupleElt n e)        = (TupleEltOp n, [e])
toHnf (SetInt e e')         = (SetIntOp, [e,e'])
toHnf (Bind e e')           = (BindOp, [e,e'])
toHnf (Return e)            = (ReturnOp, [e])
toHnf (Perform e)           = (PerformOp, [e])
toHnf (Deref8 e)            = (Deref8Op, [e])
toHnf (Deref32 e)           = (Deref32Op, [e])
toHnf Noop                  = (NoopOp, [])
toHnf (ReadArr e e')        = (ReadArrOp  ty, [e,e']) where Arr ty = typeOf e
toHnf (ReadMutArr e e')     = (ReadMutArrOp  ty, [e,e']) where Arr ty = typeOf e
toHnf (WriteArr e e' e'')   = (WriteArrOp ty, [e,e',e'']) where Arr ty = typeOf e
toHnf (NewArr e ty)         = (NewArrOp ty, [e])
toHnf Rand                  = (RandOp, [])
toHnf (NewRef e)            = (NewRefOp, [e])
toHnf (ReadRef e)           = (ReadRefOp, [e])
toHnf (WriteRef e e')       = (WriteRefOp, [e,e'])
toHnf (If e e' e'')         = (IfOp, [e,e',e''])
toHnf (DoProgress a b)      = (DoProgressOp, [a,b])
toHnf (Copy1DArray a b)     = (Copy1DArrayOp, [a,b])
toHnf e                     = error $ "not hnfable: " ++ show e


fromHnf :: (Op, [Exp]) -> Exp

fromHnf (AddOp, [e,e'])               = Add e e'
fromHnf (NegateOp, [e])               = Negate e
fromHnf (SubOp, [e,e'])               = Sub e e'
fromHnf (MulOp, [e,e'])               = Mul e e'
fromHnf (RecipOp, [e])                = Recip e
fromHnf (DivideOp, [e,e'])            = Divide e e'
fromHnf (PowOp, [e,e'])               = Pow e e'
fromHnf (LogBaseOp, [e,e'])           = LogBase e e'
fromHnf (SinOp, [e])                  = Sin e
fromHnf (Atan2Op, [e,e'])             = Atan2 e e'
fromHnf (CosOp, [e])                  = Cos e
fromHnf (I2fOp, [e])                  = I2f e
fromHnf (F2iOp, [e])                  = F2i e
fromHnf (B2iOp, [e])                  = B2i e
fromHnf (AbsOp, [e])                  = Abs e
fromHnf (ModOp, [e,e'])               = Mod e e'
fromHnf (DivOp, [e,e'])               = Div e e'
fromHnf (EqualOp, [e,e'])             = Equal e e'
fromHnf (LessThanOp, [e,e'])          = LessThan e e'
fromHnf (MinOp, [e,e'])               = Min e e'
fromHnf (MaxOp, [e,e'])               = Max e e'
fromHnf (AndOp, [e,e'])               = And e e'
fromHnf (OrOp, [e,e'])                = Or e e'
fromHnf (NotOp, [e])                  = Not e
fromHnf (LShiftOp, [e,e'])            = LShift e e'
fromHnf (RShiftOp, [e,e'])            = RShift e e'
fromHnf (BitAndOp, [e,e'])            = BitAnd e e'
fromHnf (BitOrOp, [e,e'])             = BitOr e e'
fromHnf (BitNotOp, [e])               = BitNot e
fromHnf (PackF32Op, [e1, e2, e3, e4]) = PackF32 e1 e2 e3 e4
fromHnf (NthF32Op, [e,e'])            = NthF32 e e'
fromHnf (MkTupleOp, es)               = MkTuple es
fromHnf (TupleEltOp n, [e])           = TupleElt n e
fromHnf (SumF32Op, [e])               = SumF32 e
fromHnf (SetIntOp, [e,e'])            = SetInt e e'
fromHnf (BindOp, [e,e'])              = Bind e e'
fromHnf (ReturnOp, [e])               = Return e
fromHnf (PerformOp, [e])	      = Perform e
fromHnf (Deref8Op, [e])               = Deref8 e
fromHnf (Deref32Op, [e])              = Deref32 e
fromHnf (NoopOp,[])                   = Noop
fromHnf (ReadArrOp _, [e,e'])         = ReadArr e e'
fromHnf (ReadMutArrOp _, [e,e'])      = ReadMutArr e e'
fromHnf (WriteArrOp _, [e,e',e''])    = WriteArr e e' e''
fromHnf (NewArrOp t, [e])             = NewArr e t
fromHnf (RandOp,[])                   = Rand
fromHnf (NewRefOp, [e])               = NewRef e
fromHnf (ReadRefOp, [e])              = ReadRef e
fromHnf (WriteRefOp,[e,e'])           = WriteRef e e'
fromHnf (IfOp, [e,e',e''])            = If e e' e''
fromHnf (DoProgressOp, [a,b])         = DoProgress a b
fromHnf (Copy1DArrayOp, [a,b])        = Copy1DArray a b

-- Could simply have Op derive Enum and use fromEnum, except for last two ops.

hashOp :: Op -> Int

hashOp AddOp          = 0
hashOp NegateOp       = 1
hashOp SubOp          = 2
hashOp MulOp          = 3
hashOp RecipOp        = 4
hashOp DivideOp       = 5
hashOp SinOp          = 6
hashOp CosOp          = 7
hashOp Atan2Op        = 8
hashOp I2fOp          = 10
hashOp F2iOp          = 11
hashOp B2iOp          = 12
hashOp ModOp          = 13
hashOp DivOp          = 14
hashOp EqualOp        = 15
hashOp LessThanOp     = 16
hashOp AndOp          = 17
hashOp OrOp           = 18
hashOp NotOp          = 19
hashOp LShiftOp       = 20
hashOp RShiftOp       = 21
hashOp BitAndOp       = 22
hashOp BitOrOp        = 23
hashOp BitNotOp       = 24
hashOp SetIntOp       = 25
hashOp Deref8Op       = 26
hashOp Deref32Op      = 27
hashOp NoopOp         = 29
hashOp ReadArrOp{}    = 30
hashOp ReadMutArrOp{} = 31
hashOp WriteArrOp{}   = 32
hashOp NewArrOp{}     = 33
hashOp PackF32Op      = 34
hashOp NthF32Op       = 35
hashOp MinOp          = 36
hashOp MaxOp          = 37
hashOp SumF32Op       = 38
hashOp NewRefOp       = 39
hashOp ReadRefOp      = 40
hashOp WriteRefOp     = 41
hashOp AbsOp          = 42
hashOp BindOp         = 43
hashOp ReturnOp       = 44
hashOp PerformOp      = 45
hashOp DoProgressOp   = 46
hashOp Copy1DArrayOp  = 47
hashOp PowOp          = 48
hashOp MkTupleOp      = 49
hashOp (TupleEltOp n) = 50 + n -- reserve 5
hashOp LogBaseOp      = 56
hashOp RandOp{}       = 57

numOps = 58 -- == last hashOp + 1


-- Add a clause if your op is *not* total
totalOp SetIntOp [e1,e2]      = e1
totalOp Deref8Op [e]          =  e  
totalOp Deref32Op [e]         = e
-- Try omitting and see if we get better hoisting.  Only sometimes.
totalOp (ReadArrOp _) [e1,e2] =  e2
totalOp _ _                   = []

opArity :: Op -> Int
opArity op =
  case op of
    AddOp       -> 2
    NegateOp    -> 1
    SubOp       -> 2
    MulOp       -> 2
    RecipOp     -> 1
    DivideOp    -> 2
    PowOp       -> 2
    LogBaseOp   -> 2
    SqrtOp      -> 1
    SinOp       -> 1
    CosOp       -> 1
    Atan2Op     -> 2
    I2fOp       -> 1
    F2iOp       -> 1
    B2iOp       -> 1
    AbsOp       -> 1
    ModOp       -> 2
    DivOp       -> 2
    EqualOp     -> 2
    LessThanOp  -> 2
    MinOp       -> 2
    MaxOp       -> 2
    AndOp       -> 2
    OrOp        -> 2
    NotOp       -> 1
    LShiftOp    -> 2
    BitAndOp    -> 2
    BitOrOp     -> 2
    BitNotOp    -> 1
    RShiftOp    -> 2
    PackF32Op   -> 4
    NthF32Op    -> 2
    SumF32Op    -> 1
    SetIntOp    -> 2
    BindOp      -> 2
    ReturnOp    -> 1
    PerformOp   -> 1
    Deref8Op    -> 1
    Deref32Op   -> 1
    NoopOp      -> 0
    ReadArrOp{}  -> 2
    ReadMutArrOp{} -> 2
    WriteArrOp{} -> 3
    RandOp{}     -> 0
    NewArrOp{}   -> 1
    NewRefOp     -> 1
    ReadRefOp    -> 1
    WriteRefOp   -> 2
    DoProgressOp  -> 2
    Copy1DArrayOp -> 2


-- opIsPrim op == True => emit code of the form "op arg .." (or "arg1 op
-- arg2") rather than "op(arg,..)".
-- Extend this list when adding a *primitive* operator.
opIsPrim :: Op -> Bool
opIsPrim = (`elem` primOps)

primOps = [ AddOp, NegateOp, SubOp , MulOp, DivideOp
          , I2fOp, F2iOp, B2iOp , ModOp, DivOp
          , EqualOp, LessThanOp
          , AndOp, OrOp, NotOp
          , LShiftOp, BitAndOp, BitOrOp, BitNotOp
          , RShiftOp
          ]

maxPrec, apPrec :: Int
maxPrec = 11
apPrec  = 10

-- C/C++ precedence levels (which are compatible order-wise with Haskell's
-- precedence levels.)
opPrec :: Op -> Maybe Int
opPrec op =
   case op of
    NotOp      -> Just maxPrec
    BitNotOp   -> Just maxPrec
    NegateOp   -> Just maxPrec
    RecipOp    -> Just maxPrec
    I2fOp      -> Just apPrec
    F2iOp      -> Just apPrec
    B2iOp      -> Just apPrec
    AbsOp      -> Just apPrec
    PowOp      -> Just 9
    MulOp      -> Just 8
    DivideOp   -> Just 8
    DivOp      -> Just 8
    ModOp      -> Just 8
    AddOp      -> Just 7
    SubOp      -> Just 7
    LShiftOp   -> Just 6
    RShiftOp   -> Just 6
    EqualOp    -> Just 5
    EqualOp    -> Just 5
    LessThanOp -> Just 5
    BitAndOp   -> Just 4
    BitOrOp    -> Just 3
    AndOp      -> Just 2
    OrOp       -> Just 1
    _          -> Nothing

isOpRAssoc :: Op -> Bool -- True => Right assoc.
isOpRAssoc op =
  case op of
    NotOp    -> True
    BitNotOp -> True
    PowOp    -> True
    NegateOp -> True
    I2fOp    -> False  -- (int)(char)a ==> (int)((char)a)
    AndOp    -> False  -- Note: for Haskell, && and || are right assoc.
    OrOp     -> False
    _        -> False

opAssoc :: Op -> Assoc
opAssoc op
 | isOpRAssoc op = RAssoc
 | otherwise     = LAssoc

opInfo :: Exp -> (Int, Assoc)
opInfo exp = (fromMaybe 0 (opPrec op), opAssoc op)
 where
  (op, _) = toHnf exp

freevars   :: Exp -> [Id]
freevars e = nub (fv e [])

fv (LitInt n)         = id
fv (LitDouble f)      = id
fv (LitBool b)        = id
fv (ImportData _)     = id
fv (Var n t)           = (n:)
fv (Lambda x ty e)     = \xs -> filter (/=x) (fv e []) ++ xs
fv (Let x e1 e2)       = fv e1 . (\ xs -> filter (/=x) (fv e2 []) ++ xs) 
fv (If e1 e2 e3)       = fv  e1 . fv  e2 . fv  e3
fv (Loop x e1 e2)      = \xs -> filter (/=x) (fv  e1 (fv  e2 [])) ++ xs
fv (MkArr e1 x e2)     = \xs -> filter (/=x) (fv  e1 (fv  e2 [])) ++ xs
fv appl                = foldr (.) id (map fv args)
                          where (head,args) = toHnf appl

-- TODO: reconsider Loop and MkArr cases.  Only e2 is in the scope of x.
-- Also, rework fv for memoization.

-- siz returns the size of its argument
-- One use is to fully evaluate a syntactic structure, for timing tests.

siz :: Exp -> Int

siz = memo siz'
 where
   siz' (LitInt n)        = n `seq` 1
   siz' (LitDouble f)     = f `seq` 1
   siz' (LitBool b)       = b `seq` 1
   siz' (ImportData _)    = 1
   siz' (Var n t)         = n `seq` t `seq` 1
   siz' (Lambda x ty e)   = 1 + siz e
   siz' (Let x e1 e2)     = 1 + siz e1 + siz e2
   siz' (If e1 e2 e3)     = 1 + siz e1 + siz e2 + siz e3
   siz' (Loop x e1 e2)    = 1 + siz e1 + siz e2
   siz' (MkArr e1 x e2)   = 1 + siz e1 + siz e2
   siz' appl              = 1 + sum (map siz args)
                           where (head,args) = toHnf appl


-- substitute occurrences of a variable. Note: not memoized.  Be sure to
-- use it only after CSE to avoid massive loss of representation sharing.
-- (Or maybe we should memoize.)
substVar :: Id -> Exp -> Exp -> Exp
substVar i e1 e2 =
  let sub = substVar i e1 in
  case e2 of
    Var x _ | x == i    -> e1
            | otherwise -> e2
    Lambda v t e     -> Lambda v t (sub e)
    Let v rhs e      -> Let v (sub rhs) (sub e)
    If c t f         -> If (sub c)
    			   (sub t)
    			   (sub f)
    Loop l b1 b2     -> Loop l (sub b1) (sub b2)
    MkArr a1 v a2    -> MkArr  (sub a1) v (sub a2)
    LitInt{}         -> e2
    LitDouble{}      -> e2
    LitBool{}        -> e2
    ImportData{}     -> e2
    e -> 
      let (op, args) = toHnf e2 in
      fromHnf (op, map sub args)


-- Testing 
{-
dt1 = cfold1 not (LitBool True)
dt2 = cfold1 not (Var "b" Bool)

dt3 = notE (Var "b" Bool)
dt4 = notE (LitBool True)
dt5 = notE dt3
dt6 = notE (ifE (Var "c" Bool) (Var "a" Bool) (Var "b" Bool))
dt7 = notE (ifE (Var "c" Bool) dt3 dt4)

dt8  = LitInt 0 + Var "a" Int
dt9  = Var "a" Int + LitInt 0
dt10 = LitDouble 0 + Var "a" Int
dt11 = 0 + Var "a" Int
dt12 = Var "a" Double - Var "a" Double
-}
